#!/usr/bin/env python3

# ROS libraries
import rospy
from std_msgs.msg import UInt32MultiArray, MultiArrayDimension, String
from geometry_msgs.msg import Pose2D, Twist, PoseStamped, TransformStamped
from tf2_geometry_msgs import PoseStamped
from sensor_msgs.msg import LaserScan, Imu, Image, PointCloud2
from nav_msgs.msg import Odometry
from gazebo_msgs.msg import ModelStates
from people_detection_msgs.msg import PeoplePose
from visualization_msgs.msg import Marker
from tf.transformations import euler_from_quaternion
import laser_geometry.laser_geometry as lg
import tf2_ros

import configparser
import numpy
import math
from math import sin, cos, pi
import random
from cv_bridge import CvBridge
import torch
from facenet_pytorch import MTCNN, InceptionResnetV1

# Pre-defined global variables
ROBOT_NAME = "l"
OTHER_ROBOTS = ["a","b","c","d","e","f","g","h","i","j","k"]
ROBOT_TO_ID = {
    "a" : 0,
    "b" : 1,
    "c" : 2,
    "d" : 3,
    "e" : 4,
    "f" : 5,
    "g" : 6,
    "h" : 7,
    "i" : 8,
    "j" : 9,
    "k" : 10,
    "l" : 11
}
ROBOT_NAME_CSLAM = "r11"
CONFIG_PATH = "dataset/config/scenario_loop_mini_mini.ini"

MAX_NB_ROBOTS = 0
RESULTS_DIR_PATH = "dataset/results/"
DETECTION_FILENAME = "detection_from_"

lp = lg.LaserProjection()
tf_buffer = None
tf_listener = None
tf_broadcaster = None

_model_index = 0       # New line to be added
_model_found = False   # New line to be added
_truth_pose = PoseStamped()
_cmd_vel = Twist()
_scan_ranges = [10]
_scan_low_ranges = [10]
OBSTACLE_SENSITIVITY = 0.3
ANGLE_MIN = -3.140000104904175
ANGLE_MAX = 3.140000104904175
ANGLE_INCREMENT = 0.005814814940094948
MAX_SPEED = 0.6
OBSTACLE_AVOIDANCE = 0 # 0 is for random turn, 1 for vector field
MIN_NB_TURNS = 10
MAX_NB_TURNS = 25
TURNING_SPEED_FACTOR = 4
_turning = False
_steps = 0
_direction = 0
LEFT = 1
RIGHT = -1
random.seed()


_neighbors = []
COMM_RANGE = 2 # max distance to allow robot communication

_first_pose_received = False
_first_msg_received = False
_current_pose_stamped = PoseStamped()
_new_position_in_robot_frame = PoseStamped()
#_current_pose_stamped.pose.position.x = 12
#_current_pose_stamped.pose.position.y = 12
#_current_pose_stamped.header.stamp.secs = 25
#_current_pose_stamped.header.frame_id = "robot_b_base_link"
#test_stamped = PoseStamped()
#test_stamped.pose.position.x = 9
#test_stamped.pose.position.y = 5
#test_stamped.header.stamp.secs = 42
#test_stamped.header.frame_id = "robot_b_base_link"

_people_detected = {}

#test_people_detected = {
#    1 : test_stamped,
#    3 : test_stamped
#}
#UPDATE_MIN_DIST = 2 # minimum distance to update people position
TIME_THRESHOLD = 5


pub_scan = None
pub_imu = None
pub_people = None
pub_marker = None
pub_odom_cslam = None

_marker = Marker()

PATH_TO_WEIGHTS = "dataset/resnet_weights.pth"

_last_odom_msg = Odometry()


##############################################################
##############################################################
# Practical session P1.4
# The position and orientation of the robot is given in the
# variable _trhuth_pose
# Your task is to design the path planning algorithm that
# drives the robot to the waypoints signaled in the map, and
# in the shortest possible time.
# To do so, you must drive the robot by checking its position and orientation, and
# send velocity commands accordingly.
# you must set the output _cmd_vel with the proper velocity values.


# You can create 2D pose structures such as
# _waypoint_1 = Pose2D()
# _and assign value to its x and y coordinates, and theta orientation:
# _waypoint_1.x = 1.0 # (m)
# _waypoint_1.y = 1.0 # (m)
# _waypoint_1.theta = 1.0 # (rad)[-pi,pi],
# If you require these or other class of global variables, add them here
# and declare them in the function process_data()



# Insert your code inside the function process_data
def process_data():
    # You can acces the value of the following variables:
    # _truth_pose.x , for the x coordinate of the position of the robot
    # _truth_pose.y , for the y coordinate of the position of the robot
    # _truth_pose.theta, for the orientation of the robot
    # The position and orientation of the robot is computed with
    # respect to the center of the field, in the position (0,0).
    global _truth_pose, _cmd_vel, _scan_low_ranges, OBSTACLE_SENSITIVITY, ANGLE_INCREMENT, ANGLE_MIN, MAX_SPEED, _turning, _steps, _direction, LEFT, RIGHT, OBSTACLE_AVOIDANCE, MIN_NB_TURNS, MAX_NB_TURNS, TURNING_SPEED_FACTOR

    speedVec = to_cartesian(1,0)
    for i in range(len(_scan_low_ranges)):
        if _scan_low_ranges[i] < OBSTACLE_SENSITIVITY:
            speedVec = add_vec(speedVec,to_cartesian(-1, ANGLE_MIN + (285+i)*ANGLE_INCREMENT)) 
    speedVec = norm(speedVec)
    speedVec = to_polar(speedVec)
    if speedVec[1] < 0:
        speedVec[1] += 2*math.pi
    
    if OBSTACLE_AVOIDANCE == 0:

        if not _turning:
            if speedVec[1] > 0 and speedVec[1] <= math.pi:
                _turning = True
                _direction = LEFT
                _steps = random.randint(1,10)
            elif speedVec[1] > math.pi and speedVec[1] < 2*math.pi:
                _turning = True
                _direction = RIGHT
                _steps = random.randint(MIN_NB_TURNS,MAX_NB_TURNS)
            linear = MAX_SPEED
            angular = 0
        else:
            if _steps > 0:
                _steps -= 1
            else:
                _turning = False
                _direction = 0
            linear = 0
            angular = TURNING_SPEED_FACTOR*_direction*MAX_SPEED

    elif OBSTACLE_AVOIDANCE == 1:

        # Outputs: replace the value 0.0 with your output
        # linear velocity, [-3.0,3.0] (+-1.5 m/s)
        linear = MAX_SPEED*math.cos(speedVec[1])
        if linear < 0:
            linear = 0

        # angular velocity, [-6.0,6.0] (+-3.0 rad/s)
        if speedVec[1] == 0 or speedVec[1] == 2*math.pi:
            angular = 0
        elif speedVec[1] <= math.pi:
            angular = 2*MAX_SPEED
        else:
            angular= -2*MAX_SPEED

    else:

        print("Error: No obstacle avoidance corresponds to parameter value: "+OBSTACLE_AVOIDANCE)
        linear = 0
        angular = 0


    _cmd_vel.linear.x = linear
    _cmd_vel.angular.z = angular
##############################################################
##############################################################

def in_range(position):
    global _truth_pose, COMM_RANGE
    distance = math.sqrt(((position.x-_truth_pose.pose.position.x)**2)+((position.y-_truth_pose.pose.position.y)**2))
    return distance <= COMM_RANGE

# Define callbacks and functions to process incoming messages
def gazebo_callback(msg):
    global _truth_pose, _model_index, _model_found, _neighbors, _first_msg_received, _first_pose_received, _current_pose_stamped, tf_broadcaster, _new_position_in_robot_frame

    if _model_found == False:
        _model_index = msg.name.index('robot_'+ROBOT_NAME)
        _model_found = True

    if not _first_msg_received:
        _first_msg_received = True
        _first_pose_received = True
        # Define the robot's initial position in the world frame
        robot_initial_pose = TransformStamped()
        robot_initial_pose.header.stamp = rospy.Time.now()
        robot_initial_pose.header.frame_id = "world"  # Main frame in the simulation
        robot_initial_pose.child_frame_id = "robot_"+ROBOT_NAME+"_frame"  # Robot's frame
        robot_initial_pose.transform.translation.x = msg.pose[_model_index].position.x
        robot_initial_pose.transform.translation.y = msg.pose[_model_index].position.y
        robot_initial_pose.transform.translation.z = msg.pose[_model_index].position.z
        robot_initial_pose.transform.rotation.x = msg.pose[_model_index].orientation.x
        robot_initial_pose.transform.rotation.y = msg.pose[_model_index].orientation.y
        robot_initial_pose.transform.rotation.z = msg.pose[_model_index].orientation.z
        robot_initial_pose.transform.rotation.w = msg.pose[_model_index].orientation.w
        tf_broadcaster.sendTransform(robot_initial_pose)


    _truth_pose = PoseStamped()
    _truth_pose.header.stamp = rospy.Time.now()
    _truth_pose.header.frame_id = "world"
    _truth_pose.pose.position = msg.pose[_model_index].position
    _truth_pose.pose.orientation = msg.pose[_model_index].orientation
    
    tf_buffer.can_transform("robot_"+ROBOT_NAME+"_frame", "world", rospy.Time(), rospy.Duration(1.0))
    try:
        # Get the latest transform from the world frame to the robot's frame
        transform = tf_buffer.lookup_transform("robot_"+ROBOT_NAME+"_frame", "world", rospy.Time(0), rospy.Duration(1.0))
        # Transform the new position in the world frame to the robot's frame
        # Perform the reverse transformation
        _new_position_in_robot_frame = tf_buffer.transform(_truth_pose, "robot_"+ROBOT_NAME+"_frame")
        send_odometry(copy_pose_stamped(_new_position_in_robot_frame))
        _current_pose_stamped = copy_pose_stamped(_new_position_in_robot_frame)
        
        #debug prints
        #roll = pitch = yaw = 0.0
        #orientation_q = _current_pose_stamped.pose.orientation
        #orientation_list = [orientation_q.x, orientation_q.y, orientation_q.z, orientation_q.w]
        #(roll, pitch, yaw) = euler_from_quaternion (orientation_list)
        #print("New Position in Robot Frame: ("+str(_current_pose_stamped.pose.position.x)+","+str(_current_pose_stamped.pose.position.y))
        #print("New Orientation in Robot Frame: ("+str(yaw))
        #print("Position in world Frame: ("+str(_truth_pose.x)+","+str(_truth_pose.y))
        #print("Orientation in world Frame: ("+str(_truth_pose.theta))

    except (tf2_ros.LookupException, tf2_ros.ConnectivityException, tf2_ros.ExtrapolationException) as e:
        print("Transform lookup failed: "+str(e))

    _neighbors.clear()
    for i in range(MAX_NB_ROBOTS-1):
        if in_range(msg.pose[msg.name.index('robot_'+OTHER_ROBOTS[i])].position):
            _neighbors.append(ROBOT_TO_ID[OTHER_ROBOTS[i]])


# Define callbacks and functions to process incoming messages
def image_callback(msg):
    global _first_pose_received, _people_detected, _current_pose_stamped, _truth_pose
    if _first_pose_received:
        pose = copy_pose_stamped(_current_pose_stamped)
        truth_pose = copy_pose_stamped(_truth_pose)
        bridge = CvBridge()
        img = bridge.imgmsg_to_cv2(msg, desired_encoding='passthrough')
        # Define MTCNN for face detection
        device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        mtcnn = MTCNN(keep_all=True, device=device)

        # Define Inception ResNet V1 for facial recognition and classification
        resnet = InceptionResnetV1(classify=True, num_classes=4, device=device).eval()
        resnet.load_state_dict(torch.load(PATH_TO_WEIGHTS))

        # Define Softmax to determine detection probabilities
        softmax = torch.nn.Softmax(dim=1)

        ids = forward(img, mtcnn, resnet, softmax, device)
        #for id in ids:
        #    if id in _people_detected:
        #        distance = math.sqrt(((_current_pose_stamped.pose.position.x-_people_detected[id].pose.position.x)**2)+((_current_pose_stamped.pose.position.y-_people_detected[id].pose.position.y)**2))
        #        if distance > UPDATE_MIN_DIST:
        #            _people_detected[id] = _current_pose_stamped
        #    else:
        #        _people_detected[id] = _current_pose_stamped
        updated = False
        for id in ids:
            if id != -1:
                # when testing, discovered that image is slow and position updates faster -> detection is one pose late, and uses new pose too soon 
                if id in _people_detected:# and math.sqrt(((_people_detected[id][0].pose.position.x-pose.pose.position.x)**2)+((_people_detected[id][0].pose.position.y-pose.pose.position.y)**2)) < 5:
                    _people_detected[id][0] = pose
                    _people_detected[id][1] = truth_pose
                    updated = True
                elif not id in _people_detected:
                    _people_detected[id] = []
                    _people_detected[id].append(pose)
                    _people_detected[id].append(truth_pose)
                    updated = True
        if len(_people_detected) > 0:
            send_people_detected()
            if updated:
                update_people_markers()
        

def copy_pose_stamped(pose):
    new_pose = PoseStamped()
    new_pose.header.seq = pose.header.seq
    new_pose.header.stamp = pose.header.stamp
    new_pose.header.frame_id = pose.header.frame_id
    new_pose.pose.position.x = pose.pose.position.x
    new_pose.pose.position.y = pose.pose.position.y
    new_pose.pose.position.z = pose.pose.position.z
    new_pose.pose.orientation.x = pose.pose.orientation.x
    new_pose.pose.orientation.y = pose.pose.orientation.y
    new_pose.pose.orientation.z = pose.pose.orientation.z
    new_pose.pose.orientation.w = pose.pose.orientation.w
    return new_pose

def forward(img, mtcnn, resnet, softmax, device):
	"""Return list of logits of all faces detected in img"""
	with torch.no_grad():
		boxes = mtcnn(img)
		if boxes is not None:
			logits = resnet(boxes.to(device))
			# Transform to probability per identity
			probs = softmax(logits)
			# Reject false positives
			probs = probs[(probs>0.95).any(dim=1),:]
			if probs.numel() > 0:
				return torch.argmax(probs, dim=1).tolist()

	return torch.tensor([-1]).tolist()

def send_people_detected():
    global pub_people, _people_detected
    msg = PeoplePose()
    for id, pose in _people_detected.items():
        msg.ids.append(id)
        msg.poses.append(pose[0])
        msg.truth_poses.append(pose[1])
    pub_people.publish(msg)

def people_detected_callback(msg):
    global _people_detected
    updated = False
    for i in range(len(msg.ids)):
        id = msg.ids[i]
        pose = msg.poses[i]
        truth_pose = msg.truth_poses[i]
        if id in _people_detected:
            if pose.header.stamp.secs - _people_detected[id][0].header.stamp.secs > TIME_THRESHOLD:
                _people_detected[id][0] = pose
                _people_detected[id][1] = truth_pose
                updated = True
        else:
            print("new")
            _people_detected[id] = []
            _people_detected[id].append(pose)
            _people_detected[id].append(truth_pose)
            updated = True
    if len(_people_detected) > 0 and updated:
        update_people_markers(True)

def update_people_markers(from_update=False):
    global _people_detected, _marker, pub_marker
    for id, stamped_pose in _people_detected.items():
        pose = copy_pose_stamped(stamped_pose[0])
        truth_pose = copy_pose_stamped(stamped_pose[1])
        _marker.header.frame_id = pose.header.frame_id
        _marker.header.stamp = pose.header.stamp
        _marker.id = id
        _marker.pose = pose.pose
        pub_marker.publish(_marker)
        write_results(DETECTION_FILENAME,_marker, truth_pose, from_update)


def scan_callback(msg):
    global _scan_ranges, pub_pointcloud, ROBOT_NAME_CSLAM
    _scan_ranges = msg.ranges[40:1040]
    msg.ranges = _scan_ranges
    msg.header.frame_id = "velodyne" #"/"+ROBOT_NAME_CSLAM+"/front_laser_link"
    pc2_msg = lp.projectLaser(msg)
    pub_pointcloud.publish(pc2_msg)
    #pub_scan.publish(msg)

def imu_callback(msg):
    global pub_imu, ROBOT_NAME_CSLAM
    msg.header.frame_id = "gnss" #"/"+ROBOT_NAME_CSLAM+"/imu_link"
    pub_imu.publish(msg)


def scan_low_callback(msg):
    global _scan_low_ranges
    _scan_low_ranges = msg.ranges[285:796]

def pose_estimate_callback(msg):
    global _first_pose_received, _current_pose_stamped
    if not _first_pose_received:
        _first_pose_received = True
    _current_pose_stamped = msg
    #print(_current_pose_stamped)
    
def add_vec(v1, v2):
    return [v1[0]+v2[0], v1[1]+v2[1]]

def to_cartesian(r, theta):
    return [r*math.cos(theta), r*math.sin(theta)]


def to_polar(vec):
    if vec[0] < 0:
        return [1, math.atan(vec[1]/vec[0])+math.pi]
    else:
        return [1, math.atan(vec[1]/vec[0])]

def norm(vec):
    return vec / numpy.linalg.norm(vec)

def setup_marker_msg():
    global _marker
    _marker.ns = "Person"
    # set shape, Arrow: 0; Cube: 1 ; Sphere: 2 ; Cylinder: 3
    _marker.type = Marker().SPHERE
    _marker.action = Marker().ADD
    # Set the scale of the marker
    _marker.scale.x = 1.0
    _marker.scale.y = 1.0
    _marker.scale.z = 1.0
    # Set the color
    _marker.color.r = 0.0
    _marker.color.g = 1.0
    _marker.color.b = 0.0
    _marker.color.a = 1.0

def write_results(filename, marker, truth_pose, from_update):
    timestamp = marker.header.stamp
    time = rospy.Time.now()
    id = marker.id
    x = truth_pose.pose.position.x
    y = truth_pose.pose.position.y
    file = open(RESULTS_DIR_PATH+filename, "a")
    if from_update:
        file.write(str(time.secs) + '.' + str(time.nsecs) + " : " +str(timestamp.secs) + '.' + str(timestamp.nsecs) + " : " + str(id) + " detected at ("+str(x)+","+str(y)+") by another robot\n")
    else:
        file.write(str(time.secs) + '.' + str(time.nsecs) + " : " +str(timestamp.secs) + '.' + str(timestamp.nsecs) + " : " + str(id) + " detected at ("+str(x)+","+str(y)+")\n")
    file.close()

def send_odometry(pose):
    global pub_odom_cslam, _cmd_vel
    msg = Odometry()
    msg.header = pose.header
    msg.header.frame_id = "odom"
    msg.child_frame_id = "velodyne"
    msg.pose.pose.position = pose.pose.position
    msg.pose.pose.orientation = pose.pose.orientation
    msg.twist.twist.linear = _cmd_vel.linear
    msg.twist.twist.angular = _cmd_vel.angular
    pub_odom_cslam.publish(msg)


def odom_sim_callback(msg):
    global _last_odom_msg
    _last_odom_msg = msg

def odom_cslam_callback(msg):
    global _last_odom_msg, pub_odom_cslam
    msg.pose.pose = _last_odom_msg
    pub_odom_cslam.publish(msg)



def test_callback(msg):
    print("test")

# Main

if __name__ == '__main__':

    # Add here the name of the ROS node. In ROS, node names must be unique.
    rospy.init_node('robot_'+ROBOT_NAME+'_pathplanning_node')

    config = configparser.ConfigParser()
    config.read(CONFIG_PATH)
    MAX_NB_ROBOTS = int(config['Robots']['max_nb_robots'])
    COMM_RANGE = float(config['Robots']['comm_range'])
    OBSTACLE_SENSITIVITY = float(config['Robots']['obstacle_sensitivity'])
    ANGLE_MIN = float(config['Robots']['angle_min'])
    ANGLE_MAX = float(config['Robots']['angle_max'])
    ANGLE_INCREMENT = float(config['Robots']['angle_increment'])
    MAX_SPEED = float(config['Robots']['max_speed'])
    OBSTACLE_AVOIDANCE = int(config['Robots']['obstacle_avoidance'])
    MIN_NB_TURNS = int(config['Robots']['min_nb_turns'])
    MAX_NB_TURNS = int(config['Robots']['max_nb_turns'])
    TURNING_SPEED_FACTOR = int(config['Robots']['turning_speed_factor'])
    TIME_THRESHOLD = float(config['Robots']['time_treshold'])
    PATH_TO_WEIGHTS = config['Weights']['path']
    RESULTS_DIR_PATH = config['Results']['path']
    DETECTION_FILENAME = config['Results']['detection_filename']+ROBOT_NAME_CSLAM[1:]

    setup_marker_msg()

    tf_buffer = tf2_ros.Buffer()
    tf_listener = tf2_ros.TransformListener(tf_buffer)
    tf_broadcaster = tf2_ros.StaticTransformBroadcaster()

    # Publish messages
    pub_vel = rospy.Publisher('/robot_'+ROBOT_NAME+'/cmd_vel/', Twist, queue_size = 1)
    #pub_pose = rospy.Publisher('/robot_'+ROBOT_NAME+'/mapping/truth_pose', Pose2D, queue_size = 1)
    pub_neighbors = rospy.Publisher('/'+ROBOT_NAME_CSLAM+'/cslam/list_neighbors', UInt32MultiArray, queue_size = 1)
    #pub_scan = rospy.Publisher('/'+ROBOT_NAME_CSLAM+'/scan', LaserScan, queue_size = 10)
    pub_pointcloud = rospy.Publisher('/'+ROBOT_NAME_CSLAM+'/pointcloud', PointCloud2, queue_size=1)
    pub_imu = rospy.Publisher('/'+ROBOT_NAME_CSLAM+'/imu/data', Imu, queue_size = 1)
    pub_people = rospy.Publisher('/'+ROBOT_NAME_CSLAM+'/set_people_detected', PeoplePose, queue_size = 1)
    pub_marker = rospy.Publisher("/visualization_marker", Marker, queue_size = 1)
    pub_odom_cslam = rospy.Publisher('/'+ROBOT_NAME_CSLAM+'/odom', Odometry, queue_size=1)

    # Subscribe to the topics and associate the correspondign callback functions
    sub_gazebo = rospy.Subscriber('/gazebo/model_states/', ModelStates, gazebo_callback, queue_size=1)
    sub_scan = rospy.Subscriber('/robot_'+ROBOT_NAME+'/front_laser/scan', LaserScan, scan_callback, queue_size=1)
    sub_scan_low = rospy.Subscriber('/robot_'+ROBOT_NAME+'/low_laser/scan', LaserScan, scan_low_callback, queue_size=1)
    sub_imu = rospy.Subscriber('/robot_'+ROBOT_NAME+'/imu/data', Imu, imu_callback, queue_size=1)
    sub_image = rospy.Subscriber('/'+ROBOT_NAME_CSLAM+'/image', Image, image_callback, queue_size=1)
    sub_pose_estimate = rospy.Subscriber('/'+ROBOT_NAME_CSLAM+'/cslam/current_pose_estimate', PoseStamped, pose_estimate_callback, queue_size=1)
    #sub_odom_sim = rospy.Subscriber('/robot_'+ROBOT_NAME+'/robotnik_base_control/odom', Odometry, odom_sim_callback, queue_size=1)
    #sub_odom_cslam = rospy.Subscriber('/'+ROBOT_NAME_CSLAM+'/odometry', Odometry, odom_cslam_callback, queue_size=1)
    sub_people = rospy.Subscriber('/'+ROBOT_NAME_CSLAM+'/get_people_detected', PeoplePose, people_detected_callback, queue_size=1)
    #sub_test = rospy.Subscriber('/test', String, test_callback)

    
    rate=rospy.Rate(30)

    while not rospy.is_shutdown():
        if _first_msg_received:
            process_data()
            pub_vel.publish(_cmd_vel)     # Line must be added
        #    pub_pose.publish(_truth_pose)
            data = UInt32MultiArray()
            for i in _neighbors:
                layout = MultiArrayDimension()
                layout.label = "r"+str(i)
                layout.size = 1
                layout.stride = 1
                data.layout.dim.append(layout)
            data.data = _neighbors
            pub_neighbors.publish(data)
        #_cmd_vel.linear.x = MAX_SPEED
        #_cmd_vel.angular.z = 0
        #pub_vel.publish(_cmd_vel)
        if len(_people_detected) == 4:
            print("done")
            
        rate.sleep()
